# coding=utf-8
# created by msgi on 2020/3/18 12:52 上午

import tensorflow as tf
import numpy as np
from smartnlp.layers.multi_head_attention import MultiHeadAttention
from smartnlp.layers.layer_norm import LayerNormalization
from smartnlp.layers.position_encoding import *


class EncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, n_heads, ddf, dropout_rate=0.1):
        super(EncoderLayer, self).__init__()

        self.mha = MultiHeadAttention(d_model, n_heads)
        self.ffn = point_wise_feed_forward_network(d_model, ddf)

        self.layer_norm1 = LayerNormalization(epsilon=1e-6)
        self.layer_norm2 = LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout2 = tf.keras.layers.Dropout(dropout_rate)

    def call(self, inputs, training, mask):
        # 多头注意力网络
        att_output, _ = self.mha(inputs, inputs, inputs, mask)
        att_output = self.dropout1(att_output, training=training)
        out1 = self.layer_norm1(inputs + att_output)  # (batch_size, input_seq_len, d_model)
        # 前向网络
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layer_norm1(out1 + ffn_output)  # (batch_size, input_seq_len, d_model)
        return out2


class Encoder(tf.keras.layers.Layer):
    def __init__(self, n_layers, d_model, n_heads, ddf,
                 input_vocab_size, max_seq_len, drop_rate=0.1):
        super(Encoder, self).__init__()

        self.n_layers = n_layers
        self.d_model = d_model

        self.embedding = tf.keras.layers.Embedding(input_vocab_size, d_model)
        self.pos_embedding = positional_encoding(max_seq_len, d_model)

        self.encode_layer = [EncoderLayer(d_model, n_heads, ddf, drop_rate)
                             for _ in range(n_layers)]

        self.dropout = tf.keras.layers.Dropout(drop_rate)

    def call(self, inputs, training, mark):
        seq_len = inputs.shape[1]
        word_emb = self.embedding(inputs)
        word_emb *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        emb = word_emb + self.pos_embedding[:, :seq_len, :]
        x = self.dropout(emb, training=training)
        for i in range(self.n_layers):
            x = self.encode_layer[i](x, training, mark)
        return x



from tensorflow.keras import Input, Model
from tensorflow.keras.layers import Lambda, Dense
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.preprocessing.sequence import pad_sequences
from keras_bert import load_trained_model_from_checkpoint, Tokenizer
import codecs
from sklearn.model_selection import train_test_split


class ZhTokenizer(Tokenizer):
    def _tokenize(self, text):
        R = []
        for c in text:
            if c in self._token_dict:
                R.append(c)
            elif self._is_space(c):
                R.append('[unused1]')  # space类用未经训练的[unused1]表示
            else:
                R.append('[UNK]')  # 剩余的字符是[UNK]
        return R


class BertTextClassifier:
    """
    目前示例只是拿谷歌训练好的模型来使用，不涉及训练过程（训练成本太高）
    """

    def __init__(self, config_path,
                 checkpoint_path,
                 dict_path,
                 train=False,
                 data_path=None):
        self.config_path = config_path
        self.checkpoint_path = checkpoint_path
        self.dict_path = dict_path

        if not train:
            pass
        else:
            self.data_path = data_path
            self.model = self.train()

    def train(self):
        x1_train, x2_train, y_train, x1_test, x2_test, y_test = self.preprocess()
        model = self.build_model()
        model.fit(
            [x1_train, x2_train], y_train,
            batch_size=8,
            epochs=5,
            validation_data=([x1_test, x2_test], y_test),
            verbose=1
        )
        return model

    def build_model(self):
        bert_model = load_trained_model_from_checkpoint(self.config_path, self.checkpoint_path)

        for l in bert_model.layers:
            l.trainable = True

        x1_in = Input(shape=(None,))
        x2_in = Input(shape=(None,))

        x = bert_model([x1_in, x2_in])
        x = Lambda(lambda x: x[:, 0])(x)  # 取出[CLS]对应的向量用来做分类
        p = Dense(1, activation='sigmoid')(x)

        model = Model([x1_in, x2_in], p)
        model.compile(
            loss='binary_crossentropy',
            optimizer=Adam(1e-5),  # 用足够小的学习率
            metrics=['accuracy'])
        model.summary()
        return model

    def preprocess(self):
        token_dict = {}
        with codecs.open(self.dict_path, 'r', 'utf8') as reader:
            for line in reader:
                token = line.strip()
                token_dict[token] = len(token_dict)
        tokenizer = ZhTokenizer(token_dict)

        x_data, y_data = [], []
        with open(self.data_path, 'r', encoding='utf-8') as file:
            lines = file.readlines()
        for line in lines:
            line_ = line.split('##')
            x_data.append(line_[0])
            y_data.append(int(line_[1].strip()))

        x1, x2 = [], []

        for text in x_data:
            x1_, x2_ = tokenizer.encode(first=text)
            x1.append(x1_)
            x2.append(x2_)

        x1 = pad_sequences(x1, maxlen=100)
        x2 = pad_sequences(x2, maxlen=100)

        x1_train, x1_test, x2_train, x2_test, y_train, y_test = train_test_split(x1, x2, y_data)
        return x1_train, x2_train, y_train, x1_test, x2_test, y_test



# coding:utf-8

import fasttext
from smartnlp.utils.clean_text import clean_zh_text, clean_en_text
import os
from smartnlp.utils.basic_log import Log
import logging

log = Log(logging.INFO)


class FastTextClassifier:
    """
    需要安装包：pip install fasttext
    利用fasttext来对文本进行分类
    """

    def __init__(self, model_path,
                 train=False,
                 file_path=None):
        """
        初始化
        :param file_path: 训练数据路径
        :param model_path: 模型保存路径
        """
        self.model_path = model_path
        if not train:
            self.model = self.load(self.model_path)
            assert self.model is not None, '训练模型无法获取'
        else:
            assert file_path is not None, '训练时, file_path不能为None'
            self.train_path = os.path.join(file_path, 'train.txt')
            self.test_path = os.path.join(file_path, 'test.txt')
            self.model = self.train()

    def train(self):
        """
        训练:参数可以针对性修改,进行调优
        """
        model = fasttext.supervised(self.train_path,
                                    self.model_path,
                                    label_prefix="__label__",
                                    epoch=100,
                                    dim=256,
                                    silent=False,
                                    lr=0.01)

        test_result = model.test(self.test_path)
        print('准确率: ', test_result.precision)
        return model

    def predict(self, text):
        """
        预测一条数据,由于fasttext获取的参数是列表,如果只是简单输入字符串,会将字符串按空格拆分组成列表进行推理
        :param text: 待分类的数据
        :return: 分类后的结果
        """
        if isinstance(text, list):
            output = self.model.predict(text, )
        else:
            output = self.model.predict([text], )
        print('predict:', output)
        return output

    def load(self, model_path):
        """
        加载训练好的模型
        :param model_path: 训练好的模型路径
        :return:
        """
        if os.path.exists(self.model_path + '.bin'):
            return fasttext.load_model(model_path + '.bin')
        else:
            return None


def clean(file_path):
    """
    清理文本, 然后利用清理后的文本进行训练
    """
    with open(file_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
        lines_clean = []
        for line in lines:
            line_list = line.split('__label__')
            lines_clean.append(clean_en_text(line_list[0]) + ' __label__' + line_list[1])

    with open(file_path, 'w', encoding='utf-8') as f:
        f.writelines(lines_clean)



import logging as log


class Log:
    def __init__(self, level):
        self.level = level
        log.basicConfig(format='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s',
                        level=level)
        self.log = log

    def info(self, msg):
        self.log.info(msg)

    def debug(self, msg):
        self.log.debug(msg)

    def warn(self, msg):
        self.log.warn(msg)

    def error(self, msg):
        self.log.error(msg)



# coding=utf-8
# created by msgi on 2020/3/18 12:48 上午

import tensorflow as tf


def scaled_dot_product_attention(q, k, v, mask):
    # query key 相乘获取匹配关系
    matmul_qk = tf.matmul(q, k, transpose_b=True)
    # 使用dk进行缩放
    dk = tf.cast(tf.shape(k)[-1], tf.float32)
    scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

    # 掩码
    if mask is not None:
        scaled_attention_logits += (mask * -1e9)

    # 通过soft max获取attention权重
    attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)

    # attention 乘上value
    output = tf.matmul(attention_weights, v)  # （.., seq_len_v, depth）

    return output, attention_weights


class MultiHeadAttention(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads):
        super(MultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.d_model = d_model

        # d_model 必须可以正确分为各个头
        assert d_model % num_heads == 0
        # 分头后的维度
        self.depth = d_model // num_heads

        self.wq = tf.keras.layers.Dense(d_model)
        self.wk = tf.keras.layers.Dense(d_model)
        self.wv = tf.keras.layers.Dense(d_model)

        self.dense = tf.keras.layers.Dense(d_model)

    def split_heads(self, x, batch_size):
        # 分头, 将头个数的维度 放到 seq_len 前面
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]

        # 分头前的前向网络，获取q、k、v语义
        q = self.wq(q)  # (batch_size, seq_len, d_model)
        k = self.wk(k)
        v = self.wv(v)

        # 分头
        q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)
        v = self.split_heads(v, batch_size)
        # scaled_attention.shape == (batch_size, num_heads, seq_len_v, depth)
        # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)

        # 通过缩放点积注意力层
        scaled_attention, attention_weights = scaled_dot_product_attention(
            q, k, v, mask)
        # 把多头维度后移
        scaled_attention = tf.transpose(scaled_attention, [0, 2, 1, 3])  # (batch_size, seq_len_v, num_heads, depth)

        # 合并多头
        concat_attention = tf.reshape(scaled_attention,
                                      (batch_size, -1, self.d_model))

        # 全连接重塑
        output = self.dense(concat_attention)
        return output, attention_weights



# coding:utf-8

import re


def clean_en_text(text):
    """
    清理数据,正则方式,去除标点符号等
    :param text:
    :return:
    """
    text = re.sub(r"[^A-Za-z0-9(),!?\'`]", " ", text)
    text = re.sub(r"\'s", " \'s", text)
    text = re.sub(r"\'ve", " \'ve", text)
    text = re.sub(r"n\'t", " n\'t", text)
    text = re.sub(r"\'re", " \'re", text)
    text = re.sub(r"\'d", " \'d", text)
    text = re.sub(r"\'ll", " \'ll", text)
    text = re.sub(r",", " , ", text)
    text = re.sub(r"!", " ! ", text)
    text = re.sub(r"\(", " \( ", text)
    text = re.sub(r"\)", " \) ", text)
    text = re.sub(r"\?", " \? ", text)
    text = re.sub(r"\s{2,}", " ", text)
    return text.strip().lower()


def clean_zh_text(text):
    text = re.sub(r'["\'` ?!【】\[\]./%：:&()=，,<>+_；;\-*]+', " ", text)
    return text


def clean_to_list(text):
    text = str(text)
    text = text.lower()
    # 清理数据
    text = re.sub(r"[^A-Za-z0-9^,!.\/'+-=]", " ", text)
    text = re.sub(r"what's", "what is ", text)
    text = re.sub(r"\'s", " ", text)
    text = re.sub(r"\'ve", " have ", text)
    text = re.sub(r"can't", "cannot ", text)
    text = re.sub(r"n't", " not ", text)
    text = re.sub(r"i'm", "i am ", text)
    text = re.sub(r"\'re", " are ", text)
    text = re.sub(r"\'d", " would ", text)
    text = re.sub(r"\'ll", " will ", text)
    text = re.sub(r",", " ", text)
    text = re.sub(r"\.", " ", text)
    text = re.sub(r"!", " ! ", text)
    text = re.sub(r"/", " ", text)
    text = re.sub(r"\^", " ^ ", text)
    text = re.sub(r"\+", " + ", text)
    text = re.sub(r"-", " - ", text)
    text = re.sub(r"=", " = ", text)
    text = re.sub(r"'", " ", text)
    text = re.sub(r"(\d+)(k)", r"\g<1>000", text)
    text = re.sub(r":", " : ", text)
    text = re.sub(r" e g ", " eg ", text)
    text = re.sub(r" b g ", " bg ", text)
    text = re.sub(r" u s ", " american ", text)
    text = re.sub(r"\0s", "0", text)
    text = re.sub(r" 9 11 ", "911", text)
    text = re.sub(r"e - mail", "email", text)
    text = re.sub(r"j k", "jk", text)
    text = re.sub(r"\s{2,}", " ", text)
    text = text.split()
    return text



# _*_ encoding: utf-8 _*_
import os
import fasttext


class FastTextModel:

    def __init__(self, train_file,
                 model_path,
                 model_type='skipgram'):
        """
        用facebook的fasttext训练词向量（默认skipgram方式, 如果采用cbow方式，model_type设为'cbow'）
        :param train_file: 训练的文本，文件内容是分好词的
        :param model_path: 要存储的模型路径
        """
        self.train_file = train_file
        self.model_path = model_path
        self.model_type = model_type
        self.model = self.load()
        if not self.model:
            self.model = self.train()

    # 训练模型
    def train(self):
        model = fasttext.train_unsupervised(input=self.train_file, model=self.model_type)
        model.save_model(self.model_path)
        print(model.words)
        return model

    # 返回词的向量
    def vector(self, word):
        return self.model[word]

    def get_nearest_neighbors(self, word, k):
        return self.model.get_nearest_neighbors(word, k)

    # 加载训练好的模型
    def load(self):
        if os.path.exists(self.model_path):
            return fasttext.load_model(self.model_path)
        else:
            return None



# coding=utf-8
# created by msgi on 2020/3/18 1:05 上午

import tensorflow as tf
from smartnlp.layers.encoder import Encoder
from smartnlp.layers.decoder import Decoder


class Transformer(tf.keras.Model):
    def __init__(self, n_layers, d_model, n_heads, diff,
                 input_vocab_size, target_vocab_size,
                 max_seq_len, drop_rate=0.1):
        super(Transformer, self).__init__()

        self.encoder = Encoder(n_layers, d_model, n_heads, diff,
                               input_vocab_size, max_seq_len, drop_rate)

        self.decoder = Decoder(n_layers, d_model, n_heads, diff,
                               target_vocab_size, max_seq_len, drop_rate)

        self.final_layer = tf.keras.layers.Dense(target_vocab_size)

    def call(self, inputs, targets, training, encode_padding_mask,
             look_ahead_mask, decode_padding_mask):
        encode_out = self.encoder(inputs, training, encode_padding_mask)
        print(encode_out.shape)
        decode_out, att_weights = self.decoder(targets, encode_out, training,
                                               look_ahead_mask, decode_padding_mask)
        print(decode_out.shape)
        final_out = self.final_layer(decode_out)

        return final_out, att_weights



# coding=utf-8
# created by msg on 2019/12/5 8:23 下午

import re
import unicodedata
import tensorflow as tf
from sklearn.model_selection import train_test_split


# 第一个参数指定字符串标准化的方式。
# NFC表示字符应该是整体组成(比如可能的话就使用单一编码)
# NFD表示字符应该分解为多个组合字符表示。
# unicodedata.category(chr) 把一个字符返回它在UNICODE里分类的类型
def unicode_to_ascii(s):
    return ''.join(c for c in unicodedata.normalize('NFD', s) if unicodedata.category(c) != 'Mn')


# 预处理
def preprocess_sentence(w):
    # 转小写换编码
    w = unicode_to_ascii(w.lower().strip())
    # 标点符号前后加上空格
    #  eg: "he is a boy." => "he is a boy . "
    w = re.sub(r"([?.!,¿])", r" \1 ", w)
    # 去掉多余空格
    w = re.sub(r'[" ]+', " ", w)
    # 去掉前后空格
    w = w.rstrip().strip()
    # 加开始和结束字符
    w = '<start> ' + w + ' <end>'
    return w


# 创建数据集
def parse_data(filename):
    lines = open(filename, encoding='utf-8').read().strip().split('\n')
    sentence_pairs = [line.split('\t') for line in lines]
    sentence_pairs = [(preprocess_sentence(en), preprocess_sentence(spa)) for en, spa in sentence_pairs]

    return zip(*sentence_pairs)


# 分词和词语到索引的转换
def tokenizer(lang):
    # 用空格分词
    lang_tokenizer = tf.keras.preprocessing.text.Tokenizer(num_words=None, filters='', split=' ')
    # 先分词
    lang_tokenizer.fit_on_texts(lang)
    # 再创建词到索引的映射
    tensor = lang_tokenizer.texts_to_sequences(lang)
    # 不够长度的在后边补0
    tensor = tf.keras.preprocessing.sequence.pad_sequences(tensor, padding='post')
    return tensor, lang_tokenizer


# 获取句子的最大长度
def max_length(tensor):
    return max(len(t) for t in tensor)


# 拆分训练测试数据集
def split(input_tensor, output_tensor, test_size=0.2):
    return train_test_split(input_tensor, output_tensor, test_size=test_size)


# 构建训练集
def make_dataset(input_tensor, output_tensor, batch_size=64, epochs=20, shuffle=True):
    dataset = tf.data.Dataset.from_tensor_slices((input_tensor, output_tensor))
    if shuffle:
        dataset = dataset.shuffle(30000)
    dataset = dataset.repeat(epochs).batch(batch_size, drop_remainder=True)
    return dataset


if __name__ == '__main__':
    spa_eng_path = '../../../demos/nmt/spa-eng/spa.txt'
    en_dataset, spa_dataset = parse_data(spa_eng_path)




# coding=utf-8
# created by msg on 2019/12/6 4:43 下午

import tensorflow as tf


# 编码器
class Encoder(tf.keras.Model):
    def __init__(self, vocab_size, embedding_units, encoding_units, batch_size=64):
        super().__init__()
        self.batch_size = batch_size
        self.encoding_units = encoding_units
        self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_units)
        self.gru = tf.keras.layers.GRU(self.encoding_units,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')

    def call(self, x, hidden):
        # before embedding: [batch_size, max_length, embedding_units]
        # after embedding: [batch_size, max_length, encoding_units]
        x = self.embedding(x)
        # output: [batch_size, max_length, encoding_units]
        # state: [batch_size, encoding_units]
        output, state = self.gru(x, initial_state=hidden)
        return output, state

    def initial_hidden_state(self):
        return tf.zeros((self.batch_size, self.encoding_units))


# attention层
class BahdanauAttention(tf.keras.Model):
    def __init__(self, units):
        super().__init__()
        # 三个全连接层
        self.W1 = tf.keras.layers.Dense(units)
        self.W2 = tf.keras.layers.Dense(units)
        self.V = tf.keras.layers.Dense(1)

    def call(self, decoder_hidden, encoder_outputs):
        """
        decoder_hidden: [batch_size, encoding_units]
        encoder_outputs: [batch_size, max_length, encoding_units]
        """
        # decoder_hidden_with_time_axis: [batch_size, 1, encoding_units]
        decoder_hidden_with_time_axis = tf.expand_dims(decoder_hidden, 1)
        # before V: [batch_size, max_length, units]
        # after V: [batch_size, max_length, 1]
        score = self.V(tf.nn.tanh(self.W1(encoder_outputs) + self.W2(decoder_hidden_with_time_axis)))
        # attention_weights: [batch_size, max_length, 1]
        attention_weights = tf.nn.softmax(score, axis=-1)
        # before sum: [batch_size, max_length, encoding_units]
        context_vector = attention_weights * encoder_outputs
        # after sum: [batch_size, encoding_units]
        context_vector = tf.reduce_sum(context_vector, axis=1)
        return context_vector, attention_weights


# 解码器
class Decoder(tf.keras.Model):
    def __init__(self, vocab_size, embedding_units, decoding_units, batch_size=64):
        super().__init__()
        self.batch_size = batch_size
        self.vocab_size = vocab_size
        self.decoding_units = decoding_units
        self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_units)
        self.gru = tf.keras.layers.GRU(self.decoding_units,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')
        self.fc = tf.keras.layers.Dense(vocab_size)
        self.attention = BahdanauAttention(self.decoding_units)

    def call(self, x, hidden, encoding_outputs):
        context_vector, attention_weights = self.attention(hidden, encoding_outputs)
        # before embedding: [batch_size, 1]
        # after embedding: [batch_size, 1, embedding_units]
        x = self.embedding(x)
        # combined_x: [batch_size, 1, embedding_units + encoding_units]
        combined_x = tf.concat([tf.expand_dims(context_vector, axis=1), x], axis=-1)
        # output: [batch_size, 1, decoding_units]
        # state: [batch_size, decoding_units]
        output, state = self.gru(combined_x)
        # output: [batch_size, decoding_units]
        output = tf.reshape(output, (-1, output.shape[2]))
        # output: [batch_size, vocab_size]
        output = self.fc(output)
        return output, state, attention_weights



from tensorflow.keras import Model

from smartnlp.layers.attention import Attention
from .base_classifier import TextClassifier
from tensorflow.keras.layers import Dropout, Dense, Embedding, Input, Bidirectional, LSTM


class TextRNNAttentionClassifier(TextClassifier):

    def build_model(self):
        inputs = Input(shape=(self.maxlen,))
        output = Embedding(len(self.embeddings),
                           300,
                           weights=[self.embeddings],
                           trainable=False)(inputs)
        output = Bidirectional(LSTM(150, return_sequences=True, dropout=0.25, recurrent_dropout=0.25))(output)
        output = Attention()(output)
        output = Dense(128, activation="relu")(output)
        output = Dropout(0.25)(output)
        output = Dense(1, activation="sigmoid")(output)
        model = Model(inputs=inputs, outputs=output)
        model.compile(loss='binary_crossentropy',
                      optimizer='adam',
                      metrics=['accuracy'])
        return model

    def train(self, batch_size=512, epochs=20):
        super(TextRNNAttentionClassifier, self).train(128, 3)



# coding=utf-8
# created by msgi on 2020/3/18 12:55 上午

import tensorflow as tf
from smartnlp.layers.multi_head_attention import MultiHeadAttention
from smartnlp.layers.encoder import point_wise_feed_forward_network
from smartnlp.layers.layer_norm import LayerNormalization
from smartnlp.layers.position_encoding import *


class DecoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, drop_rate=0.1):
        super(DecoderLayer, self).__init__()

        self.mha1 = MultiHeadAttention(d_model, num_heads)
        self.mha2 = MultiHeadAttention(d_model, num_heads)

        self.ffn = point_wise_feed_forward_network(d_model, dff)

        self.layer_norm1 = LayerNormalization(epsilon=1e-6)
        self.layer_norm2 = LayerNormalization(epsilon=1e-6)
        self.layer_norm3 = LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(drop_rate)
        self.dropout2 = tf.keras.layers.Dropout(drop_rate)
        self.dropout3 = tf.keras.layers.Dropout(drop_rate)

    def call(self, inputs, encode_out, training, look_ahead_mask, padding_mask):
        # masked multi-head attention
        att1, att_weight1 = self.mha1(inputs, inputs, inputs, look_ahead_mask)
        att1 = self.dropout1(att1, training=training)
        out1 = self.layer_norm1(inputs + att1)
        # multi-head attention
        att2, att_weight2 = self.mha2(encode_out, encode_out, inputs, padding_mask)
        att2 = self.dropout2(att2, training=training)
        out2 = self.layer_norm2(out1 + att2)

        ffn_out = self.ffn(out2)
        ffn_out = self.dropout3(ffn_out, training=training)
        out3 = self.layer_norm3(out2 + ffn_out)

        return out3, att_weight1, att_weight2


# import pdb
# pdb.set_trace()
class Decoder(tf.keras.layers.Layer):
    def __init__(self, n_layers, d_model, n_heads, ddf,
                 target_vocab_size, max_seq_len, drop_rate=0.1):
        super(Decoder, self).__init__()

        self.d_model = d_model
        self.n_layers = n_layers

        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_embedding = positional_encoding(max_seq_len, d_model)

        self.decoder_layers = [DecoderLayer(d_model, n_heads, ddf, drop_rate)
                               for _ in range(n_layers)]

        self.dropout = tf.keras.layers.Dropout(drop_rate)

    def call(self, inputs, encoder_out, training, look_ahead_mark, padding_mark):
        seq_len = tf.shape(inputs)[1]
        attention_weights = {}
        h = self.embedding(inputs)
        h *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        h += self.pos_embedding[:, :seq_len, :]

        h = self.dropout(h, training=training)
        # print('--------------------\n',h, h.shape)
        # 叠加解码层
        for i in range(self.n_layers):
            h, att_w1, att_w2 = self.decoder_layers[i](h, encoder_out,
                                                       training, look_ahead_mark,
                                                       padding_mark)
            attention_weights['decoder_layer{}_att_w1'.format(i + 1)] = att_w1
            attention_weights['decoder_layer{}_att_w2'.format(i + 1)] = att_w2

        return h, attention_weights



import pickle

from gensim.corpora import Dictionary
from gensim.models import LdaModel, TfidfModel
import jieba
import os


class LdaTopicModel(object):

    def __init__(self, model_path,
                 config_path,
                 train=False,
                 file_path=None):
        self.model_path = model_path
        self.config_path = config_path
        if not train:
            self.dictionary, self.tf_idf = self.load_config()
            self.model = self.load_model()
        else:
            self.file_path = file_path
            self.dictionary, self.tf_idf, self.model = self.train()

    def train(self):
        corpus = self.preprocess()
        dictionary = Dictionary(corpus)
        doc2bow = [dictionary.doc2bow(text) for text in corpus]

        tf_idf = TfidfModel(doc2bow)
        corpus_tf_idf = tf_idf[doc2bow]

        model = LdaModel(corpus_tf_idf, num_topics=2)
        return dictionary, tf_idf, model

    def save_model(self):
        self.model.save(self.model_path)

    def load_model(self):
        try:
            model = LdaModel.load(self.model_path)
        except FileNotFoundError:
            model = None
        return model

    def predict(self, text):
        line_cut = list(jieba.cut(text))
        doc2bow = self.dictionary.doc2bow(line_cut)
        corpus_tf_idf = self.tf_idf[doc2bow]
        return self.model[corpus_tf_idf]

    def save_config(self):
        with open(self.config_path, 'wb') as file:
            pickle.dump((self.dictionary, self.tf_idf), file)

    def load_config(self):
        with open(self.config_path, 'rb') as file:
            dictionary, tf_idf = pickle.load(file)
        return dictionary, tf_idf

    def preprocess(self):
        # 读取文件夹下所有的文件名
        files = os.listdir(self.file_path)
        corpus = []
        for file in files:
            dir_ = os.path.join(self.file_path, file)
            with open(dir_, 'r', encoding='utf-8') as file_:
                line = file_.read()
                line_cut = list(jieba.cut(line))
                corpus.append(line_cut)
        return corpus



# coding=utf-8
# created by msg on 2019/12/4 2:43 下午

import tensorflow as tf


def dot_product(x, kernel):
    return tf.squeeze(tf.matmul(x, tf.expand_dims(kernel, axis=-1)), axis=-1)


class Attention(tf.keras.layers.Layer):
    def __init__(self, bias=True, **kwargs):
        """
        自定义attention层
        """
        self.supports_masking = True
        self.bias = bias
        super().__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 3
        # 添加参数
        self.w = self.add_weight((input_shape[-1], input_shape[-1],),
                                 initializer='glorot_uniform',
                                 name='{}_W'.format(self.name))
        if self.bias:
            self.b = self.add_weight((input_shape[-1],),
                                     initializer='zero',
                                     name='{}_b'.format(self.name))
        self.u = self.add_weight((input_shape[-1],),
                                 initializer='glorot_uniform',
                                 name='{}_u'.format(self.name))
        super().build(input_shape)

    def call(self, x, mask=None):
        uit = dot_product(x, self.w)
        if self.bias:
            uit += self.b
        uit = tf.tanh(uit)

        ait = dot_product(uit, self.u)
        a = tf.exp(ait)

        a /= tf.cast(tf.reduce_sum(a, axis=1, keepdims=True) + tf.keras.backend.epsilon(), tf.float32)

        a = tf.expand_dims(a, axis=-1)
        weighted_input = x * a
        return tf.reduce_sum(weighted_input, axis=1)

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[-1]



# coding=utf-8
# created by msg on 2019/12/6 4:44 下午

import os
import time
from smartnlp.nmt.seq2seq.seq2seq import *
from smartnlp.nmt.seq2seq.preprocess import *


def loss_function(real, pred, loss_object):
    mask = tf.math.logical_not(tf.math.equal(real, 0))
    loss_ = loss_object(real, pred)
    mask = tf.cast(mask, dtype=loss_.dtype)
    loss_ *= mask
    return tf.reduce_mean(loss_)


def train():
    spa_eng_path = '../../../demos/nmt/spa-eng/spa.txt'
    en_dataset, spa_dataset = parse_data(spa_eng_path)
    # 西班牙语到英语的训练
    input_tensor, input_tokenizer = tokenizer(spa_dataset[:30000])
    output_tensor, output_tokenizer = tokenizer(en_dataset[:30000])

    # 拆分数据集为训练集和验证集
    input_train, input_eval, output_train, output_eval = split(input_tensor, output_tensor)
    # 构建tf.data格式
    train_dataset = make_dataset(input_train, output_train)
    eval_dataset = make_dataset(input_eval, output_eval)

    buffer_size = len(input_train)
    embedding_units = 256
    units = 1024
    batch_size = 64
    steps_per_epoch = len(input_train) // batch_size

    input_vocab_size = len(input_tokenizer.word_index) + 1
    output_vocab_size = len(output_tokenizer.word_index) + 1

    # 调用encoder
    encoder = Encoder(input_vocab_size, embedding_units, units, batch_size)
    # 调用解码器
    decoder = Decoder(output_vocab_size, embedding_units, units, batch_size)

    optimizer = tf.keras.optimizers.Adam()
    loss_object = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction='none')
    checkpoint_dir = './training_checkpoints'
    checkpoint_prefix = os.path.join(checkpoint_dir, "ckpt")
    checkpoint = tf.train.Checkpoint(optimizer=optimizer, encoder=encoder, decoder=decoder)

    # 单步训练
    def train_step(inputs, targets, enc_hidden):
        loss = 0

        # 计算梯度
        with tf.GradientTape() as tape:
            enc_output, enc_hidden = encoder(inputs, enc_hidden)
            dec_hidden = enc_hidden
            dec_inputs = tf.expand_dims([output_tokenizer.word_index['<start>']] * batch_size, 1)

            for t in range(1, targets.shape[1]):
                # 得到解码值
                predictions, dec_hidden, _ = decoder(dec_inputs, dec_hidden, enc_output)
                # 每个批次的第t个输出计算损失值
                loss += loss_function(targets[:, t], predictions, loss_object)
                dec_inputs = tf.expand_dims(targets[:, t], 1)
        # 一批数据的损失值
        batch_loss = (loss / int(targets.shape[1]))

        # 参数包括编码器参数和解码器参数
        variables = encoder.trainable_variables + decoder.trainable_variables
        # 计算梯度
        gradients = tape.gradient(loss, variables)
        # 更新权重参数
        optimizer.apply_gradients(zip(gradients, variables))

        return batch_loss

    # 10个epoch
    for epoch in range(10):
        start = time.time()
        enc_hidden = encoder.initial_hidden_state()
        total_loss = 0

        for (batch, (inputs, targets)) in enumerate(train_dataset.take(steps_per_epoch)):
            batch_loss = train_step(inputs, targets, enc_hidden)
            total_loss += batch_loss

            if batch % 100 == 0:
                print('Epoch {} Batch {} Loss {:.4f}'.format(epoch + 1, batch, batch_loss.numpy()))
        if (epoch + 1) % 2 == 0:
            checkpoint.save(file_prefix=checkpoint_prefix)

        print('Epoch {} Loss {:.4f}'.format(epoch + 1, total_loss / steps_per_epoch))
        print('Time taken for 1 epoch {} sec\n'.format(time.time() - start))


if __name__ == '__main__':
    train()



# coding:utf-8
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.feature_selection import SelectKBest, chi2
from sklearn.model_selection import train_test_split
from sklearn import svm
import pickle
import pandas as pd
from smartnlp.utils.clean_text import clean_en_text
from smartnlp.utils.basic_log import Log
import logging

log = Log(logging.INFO)


class SVMClassifier(object):
    """
    这个类,是用svm对文本进行分类.
    1. 用TF-IDF计算权重值
    2. 用卡方检验获取特征
    3. 用SVM进行分类训练
    """

    def __init__(self, model_file,
                 train_path=None,
                 train=False):
        """
        初始化参数
        :param train_path: 训练路径：数据以“x##y”的格式分隔，x为分好词的数据，y为数据标签
        :param model_file: 模型保存路径
        """
        self.model_file = model_file
        # 先读取训练好的model,如果读取不到,则重新训练
        if not train:
            self.tf_idf_model, self.chi_model, self.clf_model = self.read_model()
            assert self.tf_idf_model is not None, '训练的模型不存在,请确认后再试'
        else:
            assert train_path is not None, '训练模式下, 训练数据不能为None'
            self.train_path = train_path
            self.tf_idf_model, self.chi_model, self.clf_model = self.train_model()
            self.save_model()

    def predict(self, texts):
        """
        根据模型预测某文件的分类
        :param texts: 要分类的文本
        :return: 返回分类
        """
        texts = [clean_en_text(t) for t in texts]
        tf_vector = self.tf_idf_model.transform(texts)
        chi_vector = self.chi_model.transform(tf_vector)
        out = self.clf_model.predict(chi_vector)
        print('推理结果：', out)
        return out

    def read_model(self):
        try:
            with open(self.model_file, 'rb') as f:
                (tf_idf_model, chi_model, clf_model) = pickle.load(f)
        except FileNotFoundError:
            log.error('没有找到模型文件')
            tf_idf_model = None
            chi_model = None
            clf_model = None
        return tf_idf_model, chi_model, clf_model

    def train_model(self, test_size=0.2):
        """
        训练模型,简单地将生成的TF-IDF数据,chi提取后的特征,以及svm算法模型写入到了磁盘中
        :return: 返回训练好的模型
        """
        data_set = pd.read_csv(self.train_path,
                               sep='##',
                               encoding='utf-8',
                               header=None,
                               engine='python')
        data_set = data_set.dropna()
        chi_features, tf_idf_model, chi_model = self.__select_features(data_set)
        x_train, x_test, y_train, y_test = train_test_split(chi_features,
                                                            data_set[1],
                                                            test_size=test_size,
                                                            random_state=42)
        # 这里采用的是线性分类模型,如果采用rbf径向基模型,速度会非常慢.
        clf_model = svm.SVC(kernel='linear', verbose=True)
        print(clf_model)
        clf_model.fit(x_train, y_train)
        score = clf_model.score(x_test, y_test)
        print('测试准确率:', score)
        return tf_idf_model, chi_model, clf_model

    @staticmethod
    def __select_features(data_set):
        dataset = [clean_en_text(data) for data in data_set[0]]
        tf_idf_model = TfidfVectorizer(ngram_range=(1, 1),
                                       binary=True, 
                                       sublinear_tf=True)
        tf_vectors = tf_idf_model.fit_transform(dataset)

        # 选出前1/5的词用来做特征
        k = int(tf_vectors.shape[1] / 6)
        chi_model = SelectKBest(chi2, k=k)
        chi_features = chi_model.fit_transform(tf_vectors, data_set[1])
        print('tf-idf:\t\t' + str(tf_vectors.shape[1]))
        print('chi:\t\t' + str(chi_features.shape[1]))

        return chi_features, tf_idf_model, chi_model

    def save_model(self):
        with open(self.model_file, "wb") as file:
            pickle.dump((self.tf_idf_model, self.chi_model, self.clf_model), file)
        if file:
            file.close()



# coding=utf-8
# created by msgi on 2020/3/18 1:01 上午
import numpy as np
import tensorflow as tf


def get_angles(pos, i, d_model):
    # 这里的i等价与上面公式中的2i和2i+1
    angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(d_model))
    return pos * angle_rates


def positional_encoding(position, d_model):
    angle_rads = get_angles(np.arange(position)[:, np.newaxis],
                            np.arange(d_model)[np.newaxis, :],
                            d_model)
    # 第2i项使用sin
    sines = np.sin(angle_rads[:, 0::2])
    # 第2i+1项使用cos
    cones = np.cos(angle_rads[:, 1::2])
    pos_encoding = np.concatenate([sines, cones], axis=-1)
    pos_encoding = pos_encoding[np.newaxis, ...]

    return tf.cast(pos_encoding, dtype=tf.float32)


def point_wise_feed_forward_network(d_model, diff):
    return tf.keras.Sequential([
        tf.keras.layers.Dense(diff, activation='relu'),
        tf.keras.layers.Dense(d_model)
    ])



# coding: utf-8
from gensim.models import word2vec, KeyedVectors

from smartnlp.utils.pre_process import process_data


class GensimWord2VecModel:

    def __init__(self, train_file,
                 model_path,
                 embed_size=100,
                 vocab_path=None):
        """
        用gensim word2vec 训练词向量
        :param train_file: 分好词的文本
        :param model_path: 模型保存的路劲
        """
        self.train_file = train_file
        self.model_path = model_path
        self.vocab_path = vocab_path
        self.embed_size = embed_size
        # self.model = self.load()
        self.model = self.load_text()
        if not self.model:
            self.model = self.train()
            # self.save()
            self.save_text()

    def train(self):
        sentences = process_data(self.train_file)
        model = word2vec.Word2Vec(sentences, min_count=2, window=3, size=self.embed_size, workers=4)
        return model

    def vector(self, word):
        return self.model.wv.get_vector(word)

    def similar(self, word):
        return self.model.wv.similar_by_word(word, topn=10)

    def save(self):
        self.model.save(self.model_path)

    def save_text(self):
        self.model.wv.save_word2vec_format(self.model_path, self.vocab_path, False)

    def load(self):
        # 加载模型文件
        try:
            model = word2vec.Word2Vec.load(self.model_path)
        except FileNotFoundError:
            model = None
        return model

    def load_text(self):
        try:
            model = KeyedVectors.load_word2vec_format(self.model_path, self.vocab_path, binary=False)
        except FileNotFoundError:
            model = None
        return model



import jieba
import synonyms
import random
from random import shuffle

random.seed(2019)


def load_stopwords(stop_path):
    """
    加载停用词表
    :param stop_path:
    :return:
    """
    with open(stop_path, 'r', encoding='utf-8') as f:
        stop_words = f.readlines()
        stop_words = [stop_word.strip() for stop_word in stop_words]
    return stop_words


def get_synonyms(word):
    """
    获得词的同义词
    :param word:
    :return:
    """
    return synonyms.nearby(word)[0]


def add_word(new_words):
    """
    添加新词
    :param new_words:
    :return:
    """
    synonyms_ = []
    counter = 0
    while len(synonyms_) < 1:
        random_word = new_words[random.randint(0, len(new_words) - 1)]
        synonyms_ = get_synonyms(random_word)
        counter += 1
        if counter >= 10:
            return
    random_synonym = random.choice(synonyms_)
    random_idx = random.randint(0, len(new_words) - 1)
    new_words.insert(random_idx, random_synonym)


def swap_word(new_words):
    """
    调换位置
    :param new_words:
    :return:
    """
    random_idx_1 = random.randint(0, len(new_words) - 1)
    random_idx_2 = random_idx_1
    counter = 0
    while random_idx_2 == random_idx_1:
        random_idx_2 = random.randint(0, len(new_words) - 1)
        counter += 1
        if counter > 3:
            return new_words
    new_words[random_idx_1], new_words[random_idx_2] = new_words[random_idx_2], new_words[random_idx_1]
    return new_words


def random_deletion(words, p):
    """
    随机删除
    :param words:
    :param p:
    :return:
    """
    if len(words) == 1:
        return words

    new_words = []
    for word in words:
        r = random.uniform(0, 1)
        if r > p:
            new_words.append(word)

    if len(new_words) == 0:
        rand_int = random.randint(0, len(words) - 1)
        return [words[rand_int]]

    return new_words


def random_swap(words, n):
    """
    随机替换
    :param words:
    :param n:
    :return:
    """
    new_words = words.copy()
    for _ in range(n):
        new_words = swap_word(new_words)
    return new_words


def random_insertion(words, n):
    """
    随机添加
    :param words:
    :param n:
    :return:
    """
    new_words = words.copy()
    for _ in range(n):
        add_word(new_words)
    return new_words


def synonym_replacement(words, stopwords, n):
    """
    同义词替换
    :param words: 要替换的词
    :param stopwords: 停用词
    :param n:
    :return:
    """
    new_words = words.copy()
    random_word_list = list(set([word for word in words if word not in stopwords]))
    random.shuffle(random_word_list)
    num_replaced = 0
    for random_word in random_word_list:
        synonyms_ = get_synonyms(random_word)
        if len(synonyms_) >= 1:
            synonym = random.choice(synonyms_)
            new_words = [synonym if word == random_word else word for word in new_words]
            num_replaced += 1
        if num_replaced >= n:
            break

    sentence = ' '.join(new_words)
    new_words = sentence.split(' ')

    return new_words


class EDA:
    def __init__(self, stop_path):
        self.stopwords = load_stopwords(stop_path)

    def fit_transform(self, sentence, alpha_sr=0.1, alpha_ri=0.1, alpha_rs=0.1, p_rd=0.1, num_aug=9):
        """
        统一替换
        :param sentence:
        :param alpha_sr:
        :param alpha_ri:
        :param alpha_rs:
        :param p_rd:
        :param num_aug:
        :return:
        """
        seg_list = jieba.cut(sentence)
        seg_list = " ".join(seg_list)
        words = list(seg_list.split())
        num_words = len(words)

        augmented_sentences = []
        num_new_per_technique = int(num_aug / 4) + 1
        n_sr = max(1, int(alpha_sr * num_words))
        n_ri = max(1, int(alpha_ri * num_words))
        n_rs = max(1, int(alpha_rs * num_words))

        # 同义词替换sr
        for _ in range(num_new_per_technique):
            a_words = synonym_replacement(words, self.stopwords, n_sr)
            augmented_sentences.append(' '.join(a_words))

        # 随机插入ri
        for _ in range(num_new_per_technique):
            a_words = random_insertion(words, n_ri)
            augmented_sentences.append(' '.join(a_words))

        # 随机交换rs
        for _ in range(num_new_per_technique):
            a_words = random_swap(words, n_rs)
            augmented_sentences.append(' '.join(a_words))

        # 随机删除rd
        for _ in range(num_new_per_technique):
            a_words = random_deletion(words, p_rd)
            augmented_sentences.append(' '.join(a_words))

        shuffle(augmented_sentences)

        if num_aug >= 1:
            augmented_sentences = augmented_sentences[:num_aug]
        else:
            keep_prob = num_aug / len(augmented_sentences)
            augmented_sentences = [
                s for s in augmented_sentences if random.uniform(0, 1) < keep_prob]

        augmented_sentences.append(seg_list)

        return augmented_sentences



import pickle
import platform
from collections import Counter
import os

import numpy as np
from tensorflow.keras.layers import Embedding, Bidirectional, LSTM
from tensorflow.keras.models import Sequential
from tensorflow.keras.preprocessing.sequence import pad_sequences
from keras_contrib import losses, metrics
from keras_contrib.layers import CRF
import logging
from smartnlp.utils.basic_log import Log

log = Log(logging.INFO)


def _process_data(data, word2idx, chunk_tags, max_len=None, one_hot=False):
    if max_len is None:
        max_len = max(len(s) for s in data)
    x = [[word2idx.get(w[0].lower(), 1) for w in s] for s in data]
    y_chunk = [[chunk_tags.index(w[1]) for w in s] for s in data]

    x = pad_sequences(x, max_len)
    y_chunk = pad_sequences(y_chunk, max_len, value=-1)

    if one_hot:
        y_chunk = np.eye(len(chunk_tags), dtype='float32')[y_chunk]
    else:
        y_chunk = np.expand_dims(y_chunk, 2)
    return x, y_chunk


def _parse_data(file_path):
    if platform.system() == 'Windows':
        split_text = '\r\n'
    else:
        split_text = '\n'
    with open(file_path, 'rb') as f:
        string = f.read().decode('utf-8')
        data = [[row.split() for row in sample.split(split_text)] for sample in
                string.strip().split(split_text + split_text)]
    return data


class BiLSTMCRFNamedEntityRecognition:
    def __init__(self,
                 model_path,
                 config_path,
                 embed_dim=200,
                 rnn_units=200,
                 epochs=1,
                 batch_size=64,
                 train=False,
                 file_path=None):
        self.model_path = model_path
        self.config_path = config_path
        self.embed_dim = embed_dim
        self.rnn_units = rnn_units
        # 词性tag
        self.tags = ['O', 'B-PER', 'I-PER', 'B-LOC', 'I-LOC', "B-ORG", "I-ORG"]

        # 非训练模式，直接加载模型
        if not train:
            self.word2idx = self.__load_config()
            self.model = self.__load_model()
            assert self.model is not None, '训练模型无法获取'
        else:
            self.file_path = file_path
            self.batch_size = batch_size
            (self.train_x, self.train_y), (self.test_x, self.test_y), self.word2idx = self.__load_data()
            self.epochs = epochs
            self.model = self.train()

    # 训练
    def train(self):
        model = self.__build_model()
        model.fit(self.train_x,
                  self.train_y,
                  batch_size=self.batch_size,
                  epochs=self.epochs,
                  validation_data=[self.test_x, self.test_y])
        model.save_weights(self.model_path)
        self.__save_config()
        return model

    # 加载训练好的模型
    def __load_model(self):
        try:
            model = self.__build_model()
            model.load_weights(self.model_path)
        except FileNotFoundError:
            log.error('没有找到模型文件')
            model = None
        return model

    # 加载词表
    def __load_config(self):
        with open(self.config_path, 'rb') as config:
            word2idx = pickle.load(config)
        return word2idx

    # 保存词表
    def __save_config(self):
        with open(self.config_path, 'wb') as out:
            pickle.dump(self.word2idx, out)

    # 识别句子中的实体
    def predict(self, predict_text):
        # predict_text = ''
        sent, length = self.__preprocess_data(predict_text)
        raw = self.model.predict(sent)[0][-length:]
        result = np.argmax(raw, axis=1)
        result_tags = [self.tags[i] for i in result]

        per, loc, org = '', '', ''
        for s, t in zip(predict_text, result_tags):
            if t in ('B-PER', 'I-PER'):
                per += ' ' + s if (t == 'B-PER') else s
            if t in ('B-ORG', 'I-ORG'):
                org += ' ' + s if (t == 'B-ORG') else s
            if t in ('B-LOC', 'I-LOC'):
                loc += ' ' + s if (t == 'B-LOC') else s
        results = ['person:' + per, 'location:' + loc, 'organization:' + org]
        print(results)
        return results

    # 构造模型
    def __build_model(self):
        model = Sequential()
        model.add(Embedding(len(self.word2idx), self.embed_dim, mask_zero=True))
        model.add(Bidirectional(LSTM(self.rnn_units // 2, return_sequences=True)))
        crf = CRF(len(self.tags), sparse_target=True)
        model.add(crf)
        model.summary()
        model.compile('adam', loss=losses.crf_loss, metrics=[metrics.crf_accuracy])
        return model

    # 训练数据预处理
    def __load_data(self):
        train = _parse_data(os.path.join(self.file_path, 'train.data'))
        test = _parse_data(os.path.join(self.file_path, 'test.data'))

        # 统计每个字出现的频次
        word_counts = Counter(row[0].lower() for sample in train for row in sample)
        vocab = [w for w, f in iter(word_counts.items()) if f >= 2]
        word2idx = dict((w, i) for i, w in enumerate(vocab))

        train = _process_data(train, word2idx, self.tags)
        test = _process_data(test, word2idx, self.tags)
        return train, test, word2idx

    # 预测的时候，进行数据处理转换
    def __preprocess_data(self, data, max_len=100):
        x = [self.word2idx.get(w[0].lower(), 1) for w in data]
        length = len(x)
        x = pad_sequences([x], max_len)
        return x, length



import matplotlib.pyplot as plt


def plot(model_trained):
    # Plot accuracy
    plt.plot(model_trained.history['accuracy'])
    plt.plot(model_trained.history['val_accuracy'])
    plt.title('Model Accuracy')
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(['Train', 'Validation'], loc='upper left')
    plt.show()

    # Plot loss
    plt.plot(model_trained.history['loss'])
    plt.plot(model_trained.history['val_loss'])
    plt.title('Model Loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(['Train', 'Validation'], loc='upper right')
    plt.show()



import itertools
import logging
import os
import pickle
import time

import numpy as np
import pandas as pd
from gensim.models import KeyedVectors
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, TensorBoard
from tensorflow.keras.models import load_model
from tensorflow.keras.layers import Input, Embedding, LSTM, Bidirectional, Dense, concatenate, BatchNormalization, \
    Dropout
from tensorflow.keras.models import Model
from tensorflow.keras.preprocessing.sequence import pad_sequences
from sklearn.model_selection import train_test_split

from smartnlp.utils.basic_log import Log
from smartnlp.utils.clean_text import clean_to_list
from smartnlp.utils.plot_model_history import plot
from smartnlp.utils.set_stopwords import set_en_stopwords

log = Log(logging.INFO)


class SiameseSimilarity:
    def __init__(self, model_path,
                 config_path,
                 data_path=None,
                 embedding_file=None,
                 n_hidden=128,
                 batch_size=64,
                 epochs=10,
                 embedding_dim=300,
                 train=False):
        """
        初始化
        :param model_path: 要保存的或者已经保存的模型路径
        :param config_path: 要保存的或者已经保存的配置文件路径
        :param data_path: 存放了train.csv和test.csv的目录
        :param embedding_file: 训练好的词向量文件
        :param n_hidden: lstm隐藏层维度
        :param batch_size: 每批数目大小
        :param epochs: 
        :param train: 是否训练模式，如果是训练模式，则必须提供data_path
        """

        self.model_path = model_path
        self.config_path = config_path
        self.embedding_dim = embedding_dim
        self.n_hidden = n_hidden
        self.max_length = 500

        # 加载停用词
        self.stops = set_en_stopwords()
        if not train:
            self.embeddings, self.word_index, self.max_length = self._load_config()
            self.model = self._load_model()
        else:
            assert data_path is not None, '训练模式，训练数据必须！'
            assert embedding_file is not None, '训练模式，训练好的词向量数据必须！'
            self.data_path = data_path
            self.batch_size = batch_size
            self.epochs = epochs
            self.embedding_file = embedding_file
            self.x_train, self.y_train, self.x_val, self.y_val, self.word_index, self.max_length = self._load_data()
            self.embeddings = self._load_word2vec(self.word_index)
            self.model = self.train(call_back=True)

    def _build_model(self):
        left_input = Input(shape=(self.max_length,), dtype='int32')
        right_input = Input(shape=(self.max_length,), dtype='int32')
        embedding_layer = Embedding(len(self.embeddings),
                                    self.embedding_dim,
                                    weights=[self.embeddings],
                                    input_length=self.max_length,
                                    trainable=False)
        # Embedding
        encoded_left = embedding_layer(left_input)
        encoded_right = embedding_layer(right_input)
        # 相同的lstm网络
        shared_lstm = Bidirectional(LSTM(self.n_hidden // 2, return_sequences=True))
        shared_lstm4 = Bidirectional(LSTM(self.n_hidden // 2))

        for _ in range(3):
            encoded_left = shared_lstm(encoded_left)
        left_output = shared_lstm4(encoded_left)

        for _ in range(3):
            encoded_right = shared_lstm(encoded_right)
        right_output = shared_lstm4(encoded_right)

        # 合并后计算
        merged = concatenate([left_output, right_output])
        merged = BatchNormalization()(merged)
        merged = Dropout(0.5)(merged)
        merged = Dense(32, activation='relu')(merged)
        merged = BatchNormalization()(merged)
        merged = Dropout(0.5)(merged)
        output = Dense(1, activation='sigmoid')(merged)
        # 构造模型
        model = Model([left_input, right_input], [output])
        # Adam 优化器
        model.compile(loss='binary_crossentropy',
                      optimizer='adam',
                      metrics=['accuracy'])
        model.summary()
        return model

    def train(self, weights_only=True, call_back=False):
        model = self._build_model()

        if call_back:
            early_stopping = EarlyStopping(monitor='val_loss', patience=30)
            stamp = 'lstm_%d' % self.n_hidden
            checkpoint_dir = os.path.join(
                self.model_path, 'checkpoints/' + str(int(time.time())) + '/')
            if not os.path.exists(checkpoint_dir):
                os.makedirs(checkpoint_dir)

            bst_model_path = checkpoint_dir + stamp + '.h5'
            if weights_only:
                model_checkpoint = ModelCheckpoint(
                    bst_model_path, save_best_only=True, save_weights_only=True)
            else:
                model_checkpoint = ModelCheckpoint(
                    bst_model_path, save_best_only=True)
            tensor_board = TensorBoard(
                log_dir=checkpoint_dir + "logs/{}".format(time.time()))
            callbacks = [early_stopping, model_checkpoint, tensor_board]
        else:
            callbacks = None
        model_trained = model.fit([self.x_train['left'], self.x_train['right']],
                                  self.y_train,
                                  batch_size=self.batch_size,
                                  epochs=self.epochs,
                                  validation_data=([self.x_val['left'], self.x_val['right']], self.y_val),
                                  verbose=1,
                                  callbacks=callbacks)
        if weights_only and not call_back:
            model.save_weights(os.path.join(self.model_path, 'weights_only.h5'))
        elif not weights_only and not call_back:
            model.save(os.path.join(self.model_path, 'model.h5'))
        self._save_config()
        plot(model_trained)
        return model

    def _save_config(self):
        with open(self.config_path, 'wb') as out:
            pickle.dump((self.embeddings, self.word_index, self.max_length), out)
        if out:
            out.close()

    # 推理两个文本的相似度，大于0.5则相似，否则不相似
    def predict(self, text1, text2):
        if isinstance(text1, list) or isinstance(text2, list):
            x1 = [[self.word_index.get(word, 0) for word in clean_to_list(
                text)] for text in text1]
            x2 = [[self.word_index.get(word, 0) for word in clean_to_list(
                text)] for text in text2]
            x1 = pad_sequences(x1, maxlen=self.max_length)
            x2 = pad_sequences(x2, maxlen=self.max_length)
        else:
            x1 = [self.word_index.get(word, 0)
                  for word in clean_to_list(text1)]
            x2 = [self.word_index.get(word, 0)
                  for word in clean_to_list(text2)]
            x1 = pad_sequences([x1], maxlen=self.max_length)
            x2 = pad_sequences([x2], maxlen=self.max_length)
        # 转为词向量
        return self.model.predict([x1, x2])

    # 保存路径与加载路径相同
    def _load_model(self, weights_only=True):
        return self._load_model_by_path(self.model_path, weights_only)

    # 自定义加载的模型路径
    def _load_model_by_path(self, model_path, weights_only=True):
        try:
            if weights_only:
                model = self._build_model()
                model.load_weights(model_path)
            else:
                model = load_model(model_path)
        except FileNotFoundError:
            model = None
        return model

    # 加载word2vec词向量
    def _load_word2vec(self, word_index):
        log.info('加载训练好的词向量')
        word2vec = KeyedVectors.load_word2vec_format(
            self.embedding_file, binary=True)
        embeddings = 1 * np.random.randn(len(word_index) + 1, self.embedding_dim)
        embeddings[0] = 0

        for word, index in word_index.items():
            if word in word2vec.vocab:
                embeddings[index] = word2vec.word_vec(word)
        return embeddings

    def _load_config(self):
        log.info('加载配置文件（词向量和最大长度）')
        with open(self.config_path, 'rb') as config:
            embeddings, vocabulary, max_seq_length = pickle.load(config)
        if config:
            config.close()
        return embeddings, vocabulary, max_seq_length

    def _load_data(self, test_size=0.2):
        log.info('数据预处理...')
        # word:index和index:word
        word_index = dict()
        index_word = ['<unk>']
        questions_cols = ['question1', 'question2']

        log.info('加载数据集...')
        train_data = os.path.join(self.data_path, 'train.csv')
        test_data = os.path.join(self.data_path, 'test.csv')

        train_df = pd.read_csv(train_data)
        test_df = pd.read_csv(test_data)

        # 找到最大的句子长度
        sentences = [df[col].str.split(' ') for df in [train_df, test_df] for col in questions_cols]
        max_length = max([len(s) for ss in sentences for s in ss if isinstance(s, list)])
        # 预处理(统计并将字符串转换为索引)
        for dataset in [train_df, test_df]:
            for index, row in dataset.iterrows():
                for question_col in questions_cols:
                    question_indexes = []
                    for word in clean_to_list(row[question_col]):
                        if word in self.stops:
                            continue
                        if word not in word_index:
                            word_index[word] = len(index_word)
                            question_indexes.append(len(index_word))
                            index_word.append(word)
                        else:
                            question_indexes.append(word_index[word])
                    dataset.set_value(index, question_col, question_indexes)

        x = train_df[questions_cols]
        y = train_df['is_duplicate']
        x_train, x_val, y_train, y_val = train_test_split(x, y, test_size=test_size)

        x_train = {'left': x_train.question1, 'right': x_train.question2}
        x_val = {'left': x_val.question1, 'right': x_val.question2}

        y_train = y_train.values
        y_val = y_val.values

        for dataset, side in itertools.product([x_train, x_val], ['left', 'right']):
            dataset[side] = pad_sequences(dataset[side], maxlen=max_length)

        # 校验问题对各自数目是否正确
        assert x_train['left'].shape == x_train['right'].shape
        assert len(x_train['left']) == len(y_train)
        return x_train, y_train, x_val, y_val, word_index, max_length



# coding=utf-8
# created by msgi on 2020/3/18 12:51 上午

import tensorflow as tf


class LayerNormalization(tf.keras.layers.Layer):
    def __init__(self, epsilon=1e-6, **kwargs):
        self.eps = epsilon
        super(LayerNormalization, self).__init__(**kwargs)

    def build(self, input_shape):
        self.gamma = self.add_weight(name='gamma', shape=input_shape[-1:],
                                     initializer=tf.ones_initializer(), trainable=True)
        self.beta = self.add_weight(name='beta', shape=input_shape[-1:],
                                    initializer=tf.zeros_initializer(), trainable=True)
        super(LayerNormalization, self).build(input_shape)

    def call(self, x, **kwargs):
        mean = tf.keras.backend.mean(x, axis=-1, keepdims=True)
        std = tf.keras.backend.std(x, axis=-1, keepdims=True)
        return self.gamma * (x - mean) / (std + self.eps) + self.beta

    def compute_output_shape(self, input_shape):
        return input_shape



import jieba


# 预处理数据，分词并去除停用词，然后写回文件
def seg_to_file(file, out_file, user_dict=None, stop_dict=None):
    sentences = process_data(file, user_dict, stop_dict)
    sentences = [' '.join([l for l in line]) + '\n' for line in sentences]

    with open(out_file, 'w', encoding='utf-8') as o:
        o.writelines(sentences)


def process_data(train_file, user_dict=None, stop_dict=None):
    # 结巴分词加载自定义词典(要符合jieba自定义词典规范)
    if user_dict:
        jieba.load_userdict(user_dict)

    # 加载停用词表(每行一个停用词)
    stop_words = []
    if stop_dict:
        with open(stop_dict, 'r', encoding='utf-8') as file:
            stop_words = [stop_word.strip() for stop_word in file.readlines()]

    # 读取文件内容并分词, 去掉停用词
    with open(train_file, 'r', encoding='utf-8') as file:
        sentences = file.readlines()
        sentences = [jieba.lcut(sentence.strip()) for sentence in sentences]
        sentences = [[s for s in sentence if s not in stop_words and s.strip() != ''] for sentence in sentences]

    return sentences



import setuptools

with open("smartnlp/README.md", "r") as fh:
    long_description = fh.read()

REQUIRED_PACKAGES = [
    'h5py', 'requests'
]

setuptools.setup(
    name="smartnlp",
    version="0.0.1",
    author="msgi(慢时光)",
    author_email="mayuan120226@sina.cn",
    description="Easy-to-use and Extendable package of deep learning based smartnlp (Natural Language Processing) tools "
                "with tensorflow 2.x .",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/msgi/nlp-journey.git",
    # download_url='https://github.com/msgi/nlp-journey/tags',
    packages=setuptools.find_packages(
        exclude=["tests"]),
    python_requires=">=3.6",  # '>=3.4',  # 3.4.6
    install_requires=REQUIRED_PACKAGES,
    extras_require={
        "cpu": ["tensorflow>=2.0.0"],
        "gpu": ["tensorflow-gpu>=2.0.0"],
    },
    entry_points={
    },
    classifiers=(
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ),
    license="Apache-2.0",
    keywords=['nlp', 'natural language processing', "nlu", "natural language understanding"
                                                           'deep learning', 'tensorflow', 'keras'],
)



from tensorflow.keras import Input, Model
from tensorflow.keras.layers import Embedding, Bidirectional, CuDNNLSTM, TimeDistributed, Dense, Reshape

from smartnlp.layers.attention import Attention
from .base_classifier import TextClassifier


class TextHanClassifier(TextClassifier):

    # 对长文本比较好, 可以在长文本中截断处理，把一段作为一个sentence
    def build_model(self):
        input_word = Input(shape=(int(self.maxlen / 5),))
        x_word = Embedding(len(self.embeddings),
                           300,
                           weights=[self.embeddings],
                           trainable=False)(input_word)
        x_word = Bidirectional(CuDNNLSTM(128, return_sequences=True))(x_word)
        x_word = Attention()(x_word)
        model_word = Model(input_word, x_word)

        # Sentence part
        inputs = Input(shape=(self.maxlen,))  # (5,self.maxlen) 代表：(篇章最多包含的句子，每句包含的最大词数)
        reshape = Reshape((5, int(self.maxlen / 5)))(inputs)
        x_sentence = TimeDistributed(model_word)(reshape)
        x_sentence = Bidirectional(CuDNNLSTM(128, return_sequences=True))(x_sentence)
        x_sentence = Attention()(x_sentence)

        output = Dense(1, activation='sigmoid')(x_sentence)
        model = Model(inputs=inputs, outputs=output)
        model.compile('adam', 'binary_crossentropy', metrics=['accuracy'])
        return model

    def train(self, batch_size=512, epochs=20):
        # 比较耗费资源，笔记本GPU跑不动，只好减小batch_size
        return super(TextHanClassifier, self).train(128, 2)



from tensorflow.keras import Model
from tensorflow.keras.layers import *

from .base_classifier import TextClassifier


class TextRnnClassifier(TextClassifier):
    def __init__(self, model_path, config_path, train, vector_path):
        super(TextRnnClassifier, self).__init__(model_path, config_path, train, vector_path)

    def build_model(self):
        inputs = Input(shape=(self.maxlen,))
        x = Embedding(len(self.embeddings),
                      300,
                      weights=[self.embeddings],
                      trainable=False)(inputs)
        x = Bidirectional(LSTM(150))(x)
        x = BatchNormalization()(x)
        x = Dense(128, activation="relu")(x)
        x = Dropout(0.25)(x)
        y = Dense(1, activation="sigmoid")(x)
        model = Model(inputs=inputs, outputs=y)
        model.compile(loss='binary_crossentropy',
                      optimizer='adam',
                      metrics=['accuracy'])
        return model



from tensorflow.keras import Input, Model
from tensorflow.keras.layers import *

from .base_classifier import TextClassifier


class TextRCNNClassifier(TextClassifier):
    def build_model(self):
        inputs = Input((self.maxlen,))
        embedding = Embedding(len(self.embeddings),
                              300,
                              weights=[self.embeddings],
                              trainable=False)(inputs)
        x_context = Bidirectional(LSTM(128, return_sequences=True))(embedding)
        x = Concatenate()([embedding, x_context])
        cs = []
        for kernel_size in range(1, 5):
            c = Conv1D(128, kernel_size, activation='relu')(x)
            cs.append(c)
        pools = [GlobalAveragePooling1D()(conv) for conv in cs] + [GlobalMaxPooling1D()(c) for c in cs]
        x = Concatenate()(pools)
        output = Dense(1, activation='sigmoid')(x)
        model = Model(inputs=inputs, outputs=output)
        model.compile('adam', 'binary_crossentropy', metrics=['accuracy'])
        return model

    def train(self, batch_size=512, epochs=20):
        super(TextRCNNClassifier, self).train(128, 2)



import os
import pickle
from collections import Counter

import numpy as np
import tensorflow as tf
from gensim.models import KeyedVectors
from keras_preprocessing.sequence import pad_sequences
from sklearn.model_selection import train_test_split
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping
from tensorflow.keras.datasets import imdb
from tensorflow.keras.layers import Input, Dense, Embedding, Lambda
from tensorflow.keras.models import Model
from tensorflow.keras.models import load_model
from tensorflow.keras.utils import to_categorical

from smartnlp.utils.plot_model_history import plot


class TextClassifier:
    """
    基础文本分类器
    """

    def __init__(self, model_path,
                 config_path,
                 train=False,
                 train_file_path=None,
                 vector_path=None):
        self.model_path = model_path
        self.config_path = config_path
        if not train:
            self.word_index, self.maxlen, self.embeddings = self.load_config()
            self.model = self.load_model()
            if not self.model:
                print('模型找不到：', self.model_path)
        else:
            self.vector_path = vector_path
            self.train_file_path = train_file_path
            self.x_train, self.y_train, self.x_test, self.y_test, self.word_index = self.load_data()
            self.maxlen = self.x_train.shape[1]
            _, _, self.embeddings = self.load_config()
            if len(self.embeddings) == 0:
                self.embeddings = self.load_vector()
                self.save_config()
            self.model = self.train()
            self.save_model()

    # 全连接的一个简单的网络,仅用来作为基类测试,效果特别差
    def build_model(self):
        inputs = Input(shape=(self.maxlen,))
        x = Embedding(len(self.embeddings),
                      300,
                      weights=[self.embeddings],
                      trainable=False)(inputs)
        x = Lambda(lambda t: tf.reduce_mean(t, axis=1))(x)
        x = Dense(128, activation='relu')(x)
        x = Dense(64, activation='relu')(x)
        x = Dense(16, activation='relu')(x)
        predictions = Dense(1, activation='sigmoid')(x)
        model = Model(inputs=inputs, outputs=predictions)
        model.compile(optimizer='adam',
                      loss='binary_crossentropy',
                      metrics=['accuracy'])
        return model

    def save_model(self, weights_only=True):
        if self.model:
            if weights_only:
                self.model.save_weights(os.path.join(self.model_path, 'weights.h5'))
            else:
                self.model.save(os.path.join(self.model_path, 'model.h5'))

    def load_model(self, weights_only=True):
        try:
            if weights_only:
                model = self.build_model()
                model.load_weights(os.path.join(self.model_path, 'weights.h5'))
            else:
                model = load_model(os.path.join(self.model_path, 'model.h5'))
        except FileNotFoundError:
            model = None
        return model

    def train(self, batch_size=512, epochs=20):
        model = self.build_model()
        # early_stop配合checkpoint使用，可以得到val_loss最小的模型
        early_stop = EarlyStopping(patience=3, verbose=1)
        checkpoint = ModelCheckpoint(os.path.join(self.model_path, 'weights.{epoch:03d}-{val_loss:.3f}.h5'),
                                     verbose=1,
                                     save_best_only=True)
        history = model.fit(self.x_train,
                            self.y_train,
                            batch_size=batch_size,
                            epochs=epochs,
                            verbose=1,
                            callbacks=[checkpoint, early_stop],
                            validation_data=(self.x_test, self.y_test))
        plot(history)
        return model

    def predict(self, text):
        indices = None
        if isinstance(text, str):
            indices = [[self.word_index[t] if t in self.word_index.keys() else 0 for t in text.split()]]
        elif isinstance(text, list):
            indices = [[self.word_index[t] if t in self.word_index.keys() else 0 for t in tx.split()] for tx in text]
        if indices:
            indices = pad_sequences(indices, 500)
            return self.model.predict(indices)
        else:
            return []

    def load_config(self):
        try:
            with open(self.config_path, 'rb') as f:
                (word_index, maxlen, embeddings) = pickle.load(f)
        except FileNotFoundError:
            word_index, maxlen, embeddings = None, None, np.array([])
        return word_index, maxlen, embeddings

    def save_config(self):
        with open(self.config_path, 'wb') as f:
            pickle.dump((self.word_index, self.maxlen, self.embeddings), f)

    def summary(self):
        self.build_model().summary()

    def load_vector(self):
        return self.load_vector_en()

    def load_vector_zh(self):
        word2vec = KeyedVectors.load_word2vec_format(self.vector_path, binary=False)
        return self.load_vectors(word2vec)

    def load_vector_en(self):
        word2vec = KeyedVectors.load_word2vec_format(self.vector_path, binary=True)
        return self.load_vectors(word2vec)

    def load_vectors(self, word2vec):
        embeddings = 1 * np.random.randn(len(self.word_index) + 1, 300)
        embeddings[0] = 0
        for word, index in self.word_index.items():
            if word in word2vec.vocab:
                embeddings[index] = word2vec.word_vec(word)
        return embeddings

    def load_stopwords(self):
        from nltk.corpus import stopwords
        return stopwords.words('english')

    # 默认选用keras自带的处理好的数据来做模拟分类
    def load_data(self):
        return self.load_data_from_keras()

    def load_data_from_keras(self, maxlen=500):
        (x_train, y_train), (x_test, y_test) = imdb.load_data()

        word_index = imdb.get_word_index()

        x_train = pad_sequences(x_train, maxlen=maxlen)
        x_test = pad_sequences(x_test, x_train.shape[1])
        y_train = np.asarray(y_train).astype('float32')
        y_test = np.asarray(y_test).astype('float32')
        return x_train, y_train, x_test, y_test, word_index

    # 用自己的数据集做训练（格式：分好词的句子##标签，如：我 很 喜欢 这部 电影#pos）
    def load_data_from_scratch(self, test_size=0.2, maxlen=100):
        stopwords = self.load_stopwords()
        with open(self.train_file_path, 'r', encoding='utf-8') as file:
            lines = file.readlines()
        lines = [line.strip() for line in lines]
        lines = [line.split('##') for line in lines]
        x = [line[0] for line in lines]
        x = [line.split() for line in x]
        data = [word for xx in x for word in xx]
        y = [line[0] for line in lines]

        counter = Counter(data)
        vocab = [k for k, v in counter.items() if v >= 5]

        word_index = {k: v for v, k in enumerate(vocab)}

        max_sentence_length = max([len(words) for words in x])
        maxlen = maxlen if max_sentence_length > maxlen else max_sentence_length

        x_data = [[word_index[word] for word in words if word in word_index.keys() and word not in stopwords] for words
                  in x]
        x_data = pad_sequences(x_data, maxlen=maxlen)

        y_data = to_categorical(y)

        x_train, x_test, y_train, y_test = train_test_split(x_data, y_data, test_size=test_size)
        return x_train, y_train, x_test, y_test, word_index



from tensorflow.keras import Model
from tensorflow.keras.layers import Input, Concatenate, Dropout, Dense, Embedding, Conv1D, GlobalMaxPooling1D
from tensorflow.keras.regularizers import l2

from .base_classifier import TextClassifier


class TextCnnClassifier(TextClassifier):

    def __init__(self, model_path,
                 config_path,
                 train=False,
                 vector_path=None,
                 filter_sizes=None,
                 num_filters=256,
                 drop=0.5):
        if filter_sizes is None:
            filter_sizes = [3, 4, 5, 6]
        self.filter_sizes = filter_sizes
        self.num_filters = num_filters
        self.drop = drop
        super(TextCnnClassifier, self).__init__(model_path, config_path, train, vector_path)

    def build_model(self, input_shape=(500,)):
        inputs = Input(shape=input_shape, dtype='int32')
        embedding = Embedding(len(self.embeddings),
                              300,
                              weights=[self.embeddings],
                              trainable=False)(inputs)
        filter_results = []
        for i, filter_size in enumerate(self.filter_sizes):
            c = Conv1D(self.num_filters,
                       kernel_size=filter_size,
                       padding='valid',
                       activation='relu',
                       kernel_regularizer=l2(0.001),
                       name='conv-' + str(i + 1))(embedding)
            max_pool = GlobalMaxPooling1D(name='max-pool-' + str(i + 1))(c)
            filter_results.append(max_pool)
        concat = Concatenate()(filter_results)
        dropout = Dropout(self.drop)(concat)
        output = Dense(units=1,
                       activation='sigmoid',
                       name='dense')(dropout)
        model = Model(inputs=inputs, outputs=output)
        model.compile(optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])
        model.summary()
        return model
